@@ -15,39 +15,60 @@
 #
 # END HEADER
 
-"""This module provides the core primitives of Hypothesis, assume and given."""
+"""This module provides the core primitives of Hypothesis, such as given."""
 
 
 from __future__ import division, print_function, absolute_import
 
+import os
+import sys
 import time
 import functools
 import traceback
 from random import Random
-from collections import namedtuple
 
+import attr
+from coverage import CoverageData
+from coverage.files import canonical_filename
+from coverage.collector import Collector
+
+import hypothesis.strategies as st
 from hypothesis.errors import Flaky, Timeout, NoSuchExample, \
-    Unsatisfiable, InvalidArgument, FailedHealthCheck, \
-    UnsatisfiedAssumption, HypothesisDeprecationWarning
+    Unsatisfiable, InvalidArgument, DeadlineExceeded, MultipleFailures, \
+    FailedHealthCheck, UnsatisfiedAssumption, \
+    HypothesisDeprecationWarning
 from hypothesis.control import BuildContext
 from hypothesis._settings import settings as Settings
-from hypothesis._settings import Phase, Verbosity, HealthCheck
+from hypothesis._settings import Phase, Verbosity, HealthCheck, \
+    note_deprecation
 from hypothesis.executors import new_style_executor, \
     default_new_style_executor
 from hypothesis.reporting import report, verbose_report, current_verbosity
 from hypothesis.statistics import note_engine_for_statistics
-from hypothesis.internal.compat import str_to_bytes, getfullargspec
-from hypothesis.internal.escalation import \
+from hypothesis.internal.compat import ceil, str_to_bytes, \
+    get_type_hints, getfullargspec, encoded_filepath
+from hypothesis.internal.coverage import IN_COVERAGE_TESTS
+from hypothesis.utils.conventions import infer, not_set
+from hypothesis.internal.escalation import is_hypothesis_file, \
     escalate_hypothesis_internal_error
-from hypothesis.internal.reflection import nicerepr, arg_string, \
-    impersonate, function_digest, fully_qualified_name, \
+from hypothesis.internal.reflection import is_mock, proxies, nicerepr, \
+    arg_string, impersonate, function_digest, fully_qualified_name, \
     define_function_signature, convert_positional_arguments, \
     get_pretty_function_description
 from hypothesis.internal.conjecture.data import Status, StopTest, \
     ConjectureData
 from hypothesis.searchstrategy.strategies import SearchStrategy
 from hypothesis.internal.conjecture.engine import ExitReason, \
-    ConjectureRunner
+    ConjectureRunner, uniform, sort_key
+
+try:
+    from coverage.tracer import CFileDisposition as FileDisposition
+except ImportError:  # pragma: no cover
+    from coverage.collector import FileDisposition
+
+
+running_under_pytest = False
+global_force_seed = None
 
 
 def new_random():
@@ -67,12 +88,14 @@ def test_is_flaky(test, expected_repr):
     return test_or_flaky
 
 
-Example = namedtuple('Example', ('args', 'kwargs'))
+@attr.s()
+class Example(object):
+    args = attr.ib()
+    kwargs = attr.ib()
 
 
 def example(*args, **kwargs):
-    """Add an explicit example called with these args and kwargs to the
-    test."""
+    """A decorator which ensures a specific example is always tested."""
     if args and kwargs:
         raise InvalidArgument(
             'Cannot mix positional and keyword arguments for examples'
@@ -93,7 +116,7 @@ def example(*args, **kwargs):
 def reify_and_execute(
     search_strategy, test,
     print_example=False,
-    is_final=False,
+    is_final=False, collector=None
 ):
     def run(data):
         with BuildContext(data, is_final=is_final):
@@ -109,7 +132,15 @@ def reify_and_execute(
                 report(
                     lambda: 'Trying example: %s(%s)' % (
                         test.__name__, arg_string(test, args, kwargs)))
-            return test(*args, **kwargs)
+            if collector is None:
+                return test(*args, **kwargs)
+            else:  # pragma: no cover
+                try:
+                    collector.start()
+                    return test(*args, **kwargs)
+                finally:
+                    collector.stop()
+
     return run
 
 
@@ -154,22 +185,25 @@ def is_invalid_test(
         return invalid(
             'given must be called with at least one argument')
 
-    if (generator_arguments and (original_argspec.varargs or
-                                 original_argspec.varkw)):
+    if generator_arguments and any([original_argspec.varargs,
+                                    original_argspec.varkw,
+                                    original_argspec.kwonlyargs]):
         return invalid(
-            'varargs or keywords are not supported with positional '
-            'arguments to @given'
+            'positional arguments to @given are not supported with varargs, '
+            'varkeywords, or keyword-only arguments'
         )
 
-    if (
-        len(generator_arguments) > len(original_argspec.args)
-    ):
+    if len(generator_arguments) > len(original_argspec.args):
         return invalid((
             'Too many positional arguments for %s() (got %d but'
             ' expected at most %d') % (
                 name, len(generator_arguments),
                 len(original_argspec.args)))
 
+    if infer in generator_arguments:
+        return invalid('infer was passed as a positional argument to @given, '
+                       'but may only be passed as a keyword argument')
+
     if generator_arguments and generator_kwargs:
         return invalid(
             'cannot mix positional and keyword arguments to @given'
@@ -227,26 +261,33 @@ def execute_explicit_examples(
         # Note: Test may mutate arguments and we can't rerun explicit
         # examples, so we have to calculate the failure message at this
         # point rather than than later.
-        message_on_failure = 'Falsifying example: %s(%s)' % (
+        example_string = '%s(%s)' % (
             test.__name__, arg_string(test, arguments, example_kwargs)
         )
         try:
             with BuildContext(None) as b:
+                if settings.verbosity >= Verbosity.verbose:
+                    report('Trying example: ' + example_string)
                 test_runner(
                     None,
                     lambda data: test(*arguments, **example_kwargs)
                 )
         except BaseException:
-            traceback.print_exc()
-            report(message_on_failure)
+            report('Falsifying example: ' + example_string)
             for n in b.notes:
                 report(n)
             raise
 
 
 def fail_health_check(settings, message, label):
+    # Tell pytest to omit the body of this function from tracebacks
+    # http://doc.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
+    __tracebackhide__ = True
+
     if label in settings.suppress_health_check:
         return
+    if not settings.perform_health_check:
+        return
     message += (
         '\nSee https://hypothesis.readthedocs.io/en/latest/health'
         'checks.html for more information about this. '
@@ -259,6 +300,8 @@ def fail_health_check(settings, message, label):
 
 
 def perform_health_checks(random, settings, test_runner, search_strategy):
+    # Tell pytest to omit the body of this function from tracebacks
+    __tracebackhide__ = True
     if not settings.perform_health_check:
         return
     if not Settings.default.perform_health_check:
@@ -271,8 +314,7 @@ def perform_health_checks(random, settings, test_runner, search_strategy):
     # of loading unicode data the first time.
     data = ConjectureData(
         max_length=settings.buffer_size,
-        draw_bytes=lambda data, n, distribution:
-        distribution(health_check_random, n)
+        draw_bytes=lambda data, n: uniform(health_check_random, n)
     )
     with Settings(settings, verbosity=Verbosity.quiet):
         try:
@@ -293,8 +335,7 @@ def perform_health_checks(random, settings, test_runner, search_strategy):
         try:
             data = ConjectureData(
                 max_length=settings.buffer_size,
-                draw_bytes=lambda data, n, distribution:
-                distribution(health_check_random, n)
+                draw_bytes=lambda data, n: uniform(health_check_random, n)
             )
             with Settings(settings, verbosity=Verbosity.quiet):
                 test_runner(data, reify_and_execute(
@@ -385,22 +426,26 @@ def perform_health_checks(random, settings, test_runner, search_strategy):
 
 def get_random_for_wrapped_test(test, wrapped_test):
     settings = wrapped_test._hypothesis_internal_use_settings
+    wrapped_test._hypothesis_internal_use_generated_seed = None
 
     if wrapped_test._hypothesis_internal_use_seed is not None:
         return Random(
             wrapped_test._hypothesis_internal_use_seed)
     elif settings.derandomize:
         return Random(function_digest(test))
+    elif global_force_seed is not None:
+        return Random(global_force_seed)
     else:
-        return new_random()
+        import random
+        seed = random.getrandbits(128)
+        wrapped_test._hypothesis_internal_use_generated_seed = seed
+        return Random(seed)
 
 
 def process_arguments_to_given(
     wrapped_test, arguments, kwargs, generator_arguments, generator_kwargs,
     argspec, test, settings
 ):
-    import hypothesis.strategies as sd
-
     selfy = None
     arguments, kwargs = convert_positional_arguments(
         wrapped_test, arguments, kwargs)
@@ -412,18 +457,23 @@ def process_arguments_to_given(
         selfy = kwargs.get(argspec.args[0])
     elif arguments:
         selfy = arguments[0]
+
+    # Ensure that we don't mistake mocks for self here.
+    # This can cause the mock to be used as the test runner.
+    if is_mock(selfy):
+        selfy = None
+
     test_runner = new_style_executor(selfy)
 
     arguments = tuple(arguments)
 
-    given_specifier = sd.tuples(
-        sd.just(arguments),
-        sd.fixed_dictionaries(generator_kwargs).map(
+    search_strategy = st.tuples(
+        st.just(arguments),
+        st.fixed_dictionaries(generator_kwargs).map(
             lambda args: dict(args, **kwargs)
         )
     )
 
-    search_strategy = given_specifier
     if selfy is not None:
         search_strategy = WithRunner(search_strategy, selfy)
 
@@ -432,24 +482,262 @@ def process_arguments_to_given(
     return arguments, kwargs, test_runner, search_strategy
 
 
+def skip_exceptions_to_reraise():
+    """Return a tuple of exceptions meaning 'skip this test', to re-raise.
+
+    This is intended to cover most common test runners; if you would
+    like another to be added please open an issue or pull request.
+
+    """
+    import unittest
+    # This is a set because nose may simply re-export unittest.SkipTest
+    exceptions = set([unittest.SkipTest])
+
+    try:  # pragma: no cover
+        from unittest2 import SkipTest
+        exceptions.add(SkipTest)
+    except ImportError:
+        pass
+
+    try:  # pragma: no cover
+        from pytest.runner import Skipped
+        exceptions.add(Skipped)
+    except ImportError:
+        pass
+
+    try:  # pragma: no cover
+        from nose import SkipTest as NoseSkipTest
+        exceptions.add(NoseSkipTest)
+    except ImportError:
+        pass
+
+    return tuple(sorted(exceptions, key=str))
+
+
+exceptions_to_reraise = skip_exceptions_to_reraise()
+
+
+def new_given_argspec(original_argspec, generator_kwargs):
+    """Make an updated argspec for the wrapped test."""
+    new_args = [a for a in original_argspec.args if a not in generator_kwargs]
+    new_kwonlyargs = [a for a in original_argspec.kwonlyargs
+                      if a not in generator_kwargs]
+    annots = {k: v for k, v in original_argspec.annotations.items()
+              if k in new_args + new_kwonlyargs}
+    annots['return'] = None
+    return original_argspec._replace(
+        args=new_args, kwonlyargs=new_kwonlyargs, annotations=annots)
+
+
+HUNG_TEST_TIME_LIMIT = 5 * 60
+
+
+ROOT = os.path.dirname(__file__)
+
+STDLIB = os.path.dirname(os.__file__)
+
+
+def hypothesis_check_include(filename):  # pragma: no cover
+    if is_hypothesis_file(filename):
+        return False
+    return filename.endswith('.py')
+
+
+def escalate_warning(msg, slug=None):  # pragma: no cover
+    if slug is not None:
+        msg = '%s (%s)' % (msg, slug)
+    raise AssertionError(
+        'Unexpected warning from coverage: %s' % (msg,)
+    )
+
+
+class Arc(object):
+    __slots__ = ('filename', 'source', 'target')
+
+    def __init__(self, filename, source, target):
+        self.filename = filename
+        self.source = source
+        self.target = target
+
+
+ARC_CACHE = {}
+
+
+def arc(filename, source, target):
+    try:
+        return ARC_CACHE[filename][source][target]
+    except KeyError:
+        result = Arc(filename, source, target)
+        ARC_CACHE.setdefault(
+            filename, {}).setdefault(source, {})[target] = result
+        return result
+
+
+in_given = False
+
+FORCE_PURE_TRACER = os.getenv('HYPOTHESIS_FORCE_PURE_TRACER') == 'true'
+
+
 class StateForActualGivenExecution(object):
 
     def __init__(self, test_runner, search_strategy, test, settings, random):
         self.test_runner = test_runner
         self.search_strategy = search_strategy
-        self.test = test
         self.settings = settings
         self.at_least_one_success = False
         self.last_exception = None
         self.repr_for_last_exception = None
-        self.falsifying_example = None
+        self.falsifying_examples = ()
+        self.__was_flaky = False
         self.random = random
+        self.__warned_deadline = False
+        self.__existing_collector = None
+        self.__test_runtime = None
+        self.__in_final_replay = False
+
+        if self.settings.deadline is None:
+            self.test = test
+        else:
+            @proxies(test)
+            def timed_test(*args, **kwargs):
+                self.__test_runtime = None
+                start = time.time()
+                result = test(*args, **kwargs)
+                runtime = (time.time() - start) * 1000
+                self.__test_runtime = runtime
+                if self.settings.deadline is not_set:
+                    if (
+                        not self.__warned_deadline and
+                        runtime >= 200
+                    ):
+                        self.__warned_deadline = True
+                        note_deprecation((
+                            'Test took %.2fms to run. In future the default '
+                            'deadline setting will be 200ms, which will '
+                            'make this an error. You can set deadline to '
+                            'an explicit value of e.g. %d to turn tests '
+                            'slower than this into an error, or you can set '
+                            'it to None to disable this check entirely.') % (
+                                runtime, ceil(runtime / 100) * 100,
+                        ))
+                elif runtime >= self.current_deadline:
+                    raise DeadlineExceeded(runtime, self.settings.deadline)
+                return result
+            self.test = timed_test
+
+        self.coverage_data = CoverageData()
+        self.files_to_propagate = set()
+
+        if settings.use_coverage and not IN_COVERAGE_TESTS:  # pragma: no cover
+            if Collector._collectors:
+                self.hijack_collector(Collector._collectors[-1])
+
+            self.collector = Collector(
+                branch=True,
+                timid=FORCE_PURE_TRACER,
+                should_trace=self.should_trace,
+                check_include=hypothesis_check_include,
+                concurrency='thread',
+                warn=escalate_warning,
+            )
+            self.collector.reset()
+
+            # Hide the other collectors from this one so it doesn't attempt to
+            # pause them (we're doing trace function management ourselves so
+            # this will just cause problems).
+            self.collector._collectors = []
+        else:
+            self.collector = None
+
+    @property
+    def current_deadline(self):
+        base = self.settings.deadline
+        if self.__in_final_replay:
+            return base
+        else:
+            return base * 1.25
+
+    def should_trace(self, original_filename, frame):  # pragma: no cover
+        disp = FileDisposition()
+        assert original_filename is not None
+        disp.original_filename = original_filename
+        disp.canonical_filename = encoded_filepath(
+            canonical_filename(original_filename))
+        disp.source_filename = disp.canonical_filename
+        disp.reason = ''
+        disp.file_tracer = None
+        disp.has_dynamic_filename = False
+        disp.trace = hypothesis_check_include(disp.canonical_filename)
+        if not disp.trace:
+            disp.reason = 'hypothesis internal reasons'
+        elif self.__existing_collector is not None:
+            check = self.__existing_collector.should_trace(
+                original_filename, frame)
+            if check.trace:
+                self.files_to_propagate.add(check.canonical_filename)
+        return disp
+
+    def hijack_collector(self, collector):  # pragma: no cover
+        self.__existing_collector = collector
+        original_save_data = collector.save_data
+
+        def save_data(covdata):
+            original_save_data(covdata)
+            if collector.branch:
+                covdata.add_arcs({
+                    filename: {
+                        arc: None
+                        for arc in self.coverage_data.arcs(filename)}
+                    for filename in self.files_to_propagate
+                })
+            else:
+                covdata.add_lines({
+                    filename: {
+                        line: None
+                        for line in self.coverage_data.lines(filename)}
+                    for filename in self.files_to_propagate
+                })
+            collector.save_data = original_save_data
+        collector.save_data = save_data
 
     def evaluate_test_data(self, data):
+        if (
+            time.time() - self.start_time >= HUNG_TEST_TIME_LIMIT
+        ):
+            fail_health_check(self.settings, (
+                'Your test has been running for at least five minutes. This '
+                'is probably not what you intended, so by default Hypothesis '
+                'turns it into an error.'
+            ), HealthCheck.hung_test)
+
         try:
-            result = self.test_runner(data, reify_and_execute(
-                self.search_strategy, self.test,
-            ))
+            if self.collector is None:
+                result = self.test_runner(data, reify_and_execute(
+                    self.search_strategy, self.test,
+                ))
+            else:  # pragma: no cover
+                # This should always be a no-op, but the coverage tracer has
+                # a bad habit of resurrecting itself.
+                original = sys.gettrace()
+                sys.settrace(None)
+                try:
+                    self.collector.data = {}
+                    result = self.test_runner(data, reify_and_execute(
+                        self.search_strategy, self.test,
+                        collector=self.collector
+                    ))
+                finally:
+                    sys.settrace(original)
+                    covdata = CoverageData()
+                    self.collector.save_data(covdata)
+                    self.coverage_data.update(covdata)
+                    for filename in covdata.measured_files():
+                        if is_hypothesis_file(filename):
+                            continue
+                        data.tags.update(
+                            arc(filename, source, target)
+                            for source, target in covdata.arcs(filename)
+                        )
             if result is not None and self.settings.perform_health_check:
                 fail_health_check(self.settings, (
                     'Tests run under @given should return None, but '
@@ -462,38 +750,68 @@ class StateForActualGivenExecution(object):
         except (
             HypothesisDeprecationWarning, FailedHealthCheck,
             StopTest,
-        ):
+        ) + exceptions_to_reraise:
             raise
-        except Exception:
+        except Exception as e:
             escalate_hypothesis_internal_error()
-            self.last_exception = traceback.format_exc()
-            verbose_report(self.last_exception)
-            data.mark_interesting()
+            data.__expected_traceback = traceback.format_exc()
+            data.__expected_exception = e
+            verbose_report(data.__expected_traceback)
+
+            error_class, _, tb = sys.exc_info()
+
+            origin = traceback.extract_tb(tb)[-1]
+            filename = origin[0]
+            lineno = origin[1]
+            data.mark_interesting((error_class, filename, lineno))
 
     def run(self):
+        # Tell pytest to omit the body of this function from tracebacks
+        __tracebackhide__ = True
         database_key = str_to_bytes(fully_qualified_name(self.test))
-        start_time = time.time()
+        self.start_time = time.time()
+        global in_given
         runner = ConjectureRunner(
             self.evaluate_test_data,
             settings=self.settings, random=self.random,
             database_key=database_key,
         )
-        runner.run()
+
+        if in_given or self.collector is None:
+            runner.run()
+        else:  # pragma: no cover
+            in_given = True
+            original_trace = sys.gettrace()
+            try:
+                sys.settrace(None)
+                runner.run()
+            finally:
+                in_given = False
+                sys.settrace(original_trace)
         note_engine_for_statistics(runner)
-        run_time = time.time() - start_time
-        timed_out = (
-            self.settings.timeout > 0 and
-            run_time >= self.settings.timeout
-        )
+        run_time = time.time() - self.start_time
+        timed_out = runner.exit_reason == ExitReason.timeout
         if runner.last_data is None:
             return
-        if runner.last_data.status == Status.INTERESTING:
-            self.falsifying_example = runner.last_data.buffer
-            if self.settings.database is not None:
-                self.settings.database.save(
-                    database_key, self.falsifying_example
-                )
+        if runner.interesting_examples:
+            self.falsifying_examples = sorted(
+                [d for d in runner.interesting_examples.values()],
+                key=lambda d: sort_key(d.buffer), reverse=True
+            )
         else:
+            if timed_out:
+                note_deprecation((
+                    'Your tests are hitting the settings timeout (%.2fs). '
+                    'This functionality will go away in a future release '
+                    'and you should not rely on it. Instead, try setting '
+                    'max_examples to be some value lower than %d (the number '
+                    'of examples your test successfully ran here). Or, if you '
+                    'would prefer your tests to run to completion, regardless '
+                    'of how long they take, you can set the timeout value to '
+                    'hypothesis.unlimited.'
+                ) % (
+                    self.settings.timeout, runner.valid_examples),
+                    self.settings)
             if runner.valid_examples < min(
                 self.settings.min_satisfying_examples,
                 self.settings.max_examples,
@@ -520,58 +838,113 @@ class StateForActualGivenExecution(object):
                         get_pretty_function_description(self.test),
                         runner.valid_examples,))
 
-        if self.falsifying_example is None:
+        if not self.falsifying_examples:
             return
 
-        assert self.last_exception is not None
+        flaky = 0
 
-        try:
-            with self.settings:
-                self.test_runner(
-                    ConjectureData.for_buffer(self.falsifying_example),
-                    reify_and_execute(
-                        self.search_strategy, self.test,
-                        print_example=True, is_final=True
+        self.__in_final_replay = True
+
+        for falsifying_example in self.falsifying_examples:
+            self.__was_flaky = False
+            try:
+                with self.settings:
+                    self.test_runner(
+                        ConjectureData.for_buffer(falsifying_example.buffer),
+                        reify_and_execute(
+                            self.search_strategy, self.test,
+                            print_example=True, is_final=True
+                        ))
+            except (UnsatisfiedAssumption, StopTest):
+                report(traceback.format_exc())
+                self.__flaky(
+                    'Unreliable assumption: An example which satisfied '
+                    'assumptions on the first run now fails it.'
+                )
+            except BaseException:
+                if len(self.falsifying_examples) <= 1:
+                    raise
+                report(traceback.format_exc())
+            else:
+                if (
+                    isinstance(
+                        falsifying_example.__expected_exception,
+                        DeadlineExceeded
+                    ) and self.__test_runtime is not None
+                ):
+                    report((
+                        'Unreliable test timings! On an initial run, this '
+                        'test took %.2fms, which exceeded the deadline of '
+                        '%.2fms, but on a subsequent run it took %.2f ms, '
+                        'which did not. If you expect this sort of '
+                        'variability in your test timings, consider turning '
+                        'deadlines off for this test by setting deadline=None.'
+                    ) % (
+                        falsifying_example.__expected_exception.runtime,
+                        self.settings.deadline, self.__test_runtime
                     ))
-        except (UnsatisfiedAssumption, StopTest):
-            report(traceback.format_exc())
-            raise Flaky(
-                'Unreliable assumption: An example which satisfied '
-                'assumptions on the first run now fails it.'
-            )
+                else:
+                    report(
+                        'Failed to reproduce exception. Expected: \n' +
+                        falsifying_example.__expected_traceback,
+                    )
 
-        report(
-            'Failed to reproduce exception. Expected: \n' +
-            self.last_exception,
-        )
+                filter_message = (
+                    'Unreliable test data: Failed to reproduce a failure '
+                    'and then when it came to recreating the example in '
+                    'order to print the test data with a flaky result '
+                    'the example was filtered out (by e.g. a '
+                    'call to filter in your strategy) when we didn\'t '
+                    'expect it to be.'
+                )
 
-        filter_message = (
-            'Unreliable test data: Failed to reproduce a failure '
-            'and then when it came to recreating the example in '
-            'order to print the test data with a flaky result '
-            'the example was filtered out (by e.g. a '
-            'call to filter in your strategy) when we didn\'t '
-            'expect it to be.'
-        )
+                try:
+                    self.test_runner(
+                        ConjectureData.for_buffer(falsifying_example.buffer),
+                        reify_and_execute(
+                            self.search_strategy,
+                            test_is_flaky(
+                                self.test, self.repr_for_last_exception),
+                            print_example=True, is_final=True
+                        ))
+                except (UnsatisfiedAssumption, StopTest):
+                    self.__flaky(filter_message)
+                except Flaky as e:
+                    if len(self.falsifying_examples) > 1:
+                        self.__flaky(e.args[0])
+                    else:
+                        raise
 
-        try:
-            self.test_runner(
-                ConjectureData.for_buffer(self.falsifying_example),
-                reify_and_execute(
-                    self.search_strategy,
-                    test_is_flaky(self.test, self.repr_for_last_exception),
-                    print_example=True, is_final=True
-                ))
-        except (UnsatisfiedAssumption, StopTest):
-            raise Flaky(filter_message)
+            if self.__was_flaky:
+                flaky += 1
+
+        # If we only have one example then we should have raised an error or
+        # flaky prior to this point.
+        assert len(self.falsifying_examples) > 1
+
+        if flaky > 0:
+            raise Flaky((
+                'Hypothesis found %d distinct failures, but %d of them '
+                'exhibited some sort of flaky behaviour.') % (
+                    len(self.falsifying_examples), flaky))
+        else:
+            raise MultipleFailures((
+                'Hypothesis found %d distinct failures.') % (
+                    len(self.falsifying_examples,)))
+
+    def __flaky(self, message):
+        if len(self.falsifying_examples) <= 1:
+            raise Flaky(message)
+        else:
+            self.__was_flaky = True
+            report('Flaky example! ' + message)
 
 
 def given(*given_arguments, **given_kwargs):
     """A decorator for turning a test function that accepts arguments into a
     randomized test.
 
-    This is the main entry point to Hypothesis. See the full tutorial
-    for details of its behaviour.
+    This is the main entry point to Hypothesis.
 
     """
     def run_test_with_generator(test):
@@ -587,30 +960,45 @@ def given(*given_arguments, **given_kwargs):
         if check_invalid is not None:
             return check_invalid
 
-        arguments = original_argspec.args
-
-        for name, strategy in zip(arguments[-len(generator_arguments):],
-                                  generator_arguments):
+        for name, strategy in zip(reversed(original_argspec.args),
+                                  reversed(generator_arguments)):
             generator_kwargs[name] = strategy
 
-        argspec = original_argspec._replace(
-            args=[a for a in arguments if a not in generator_kwargs],
-            defaults=None,
-        )
-        annots = {k: v for k, v in argspec.annotations.items()
-                  if k in (argspec.args + argspec.kwonlyargs)}
-        annots['return'] = None
-        argspec = argspec._replace(annotations=annots)
+        argspec = new_given_argspec(original_argspec, generator_kwargs)
 
         @impersonate(test)
         @define_function_signature(
             test.__name__, test.__doc__, argspec
         )
         def wrapped_test(*arguments, **kwargs):
+            # Tell pytest to omit the body of this function from tracebacks
+            __tracebackhide__ = True
+
+            if getattr(test, 'is_hypothesis_test', False):
+                note_deprecation((
+                    'You have applied @given to a test more than once. In '
+                    'future this will be an error. Applying @given twice '
+                    'wraps the test twice, which can be extremely slow. A '
+                    'similar effect can be gained by combining the arguments '
+                    'to the two calls to given. For example, instead of '
+                    '@given(booleans()) @given(integers()), you could write '
+                    '@given(booleans(), integers())'
+                ))
+
             settings = wrapped_test._hypothesis_internal_use_settings
 
             random = get_random_for_wrapped_test(test, wrapped_test)
 
+            if infer in generator_kwargs.values():
+                hints = get_type_hints(test)
+            for name in [name for name, value in generator_kwargs.items()
+                         if value is infer]:
+                if name not in hints:
+                    raise InvalidArgument(
+                        'passed %s=infer for %s, but %s has no type annotation'
+                        % (name, test.__name__, name))
+                generator_kwargs[name] = st.from_type(hints[name])
+
             processed_args = process_arguments_to_given(
                 wrapped_test, arguments, kwargs, generator_arguments,
                 generator_kwargs, argspec, test, settings
@@ -630,16 +1018,32 @@ def given(*given_arguments, **given_kwargs):
             ):
                 return
 
-            perform_health_checks(
-                random, settings, test_runner, search_strategy)
+            try:
+                perform_health_checks(
+                    random, settings, test_runner, search_strategy)
 
-            state = StateForActualGivenExecution(
-                test_runner, search_strategy, test, settings, random)
-            state.run()
+                state = StateForActualGivenExecution(
+                    test_runner, search_strategy, test, settings, random)
+                state.run()
+            except BaseException:
+                generated_seed = \
+                    wrapped_test._hypothesis_internal_use_generated_seed
+                if generated_seed is not None:
+                    if running_under_pytest:
+                        report((
+                            'You can add @seed(%(seed)d) to this test or run '
+                            'pytest with --hypothesis-seed=%(seed)d to '
+                            'reproduce this failure.') % {
+                                'seed': generated_seed},)
+                    else:
+                        report((
+                            'You can add @seed(%d) to this test to reproduce '
+                            'this failure.') % (generated_seed,))
+                raise
 
-        for attr in dir(test):
-            if attr[0] != '_' and not hasattr(wrapped_test, attr):
-                setattr(wrapped_test, attr, getattr(test, attr))
+        for attrib in dir(test):
+            if not (attrib.startswith('_') or hasattr(wrapped_test, attrib)):
+                setattr(wrapped_test, attrib, getattr(test, attrib))
         wrapped_test.is_hypothesis_test = True
         wrapped_test._hypothesis_internal_use_seed = getattr(
             test, '_hypothesis_internal_use_seed', None
@@ -652,6 +1056,8 @@ def given(*given_arguments, **given_kwargs):
 
 
 def find(specifier, condition, settings=None, random=None, database_key=None):
+    """Returns the minimal example from the given strategy ``specifier`` that
+    matches the predicate function ``condition``."""
     settings = settings or Settings(
         max_examples=2000,
         min_satisfying_examples=0,
